import 'dart:collection';

import 'package:fluartx/fluartx.dart';
import 'package:fluartx/src/annotations.dart';

class Collections {

  @Deprecated("use extension properity of IterableX since dart 2.6")
  static bool isNullOrEmpty(@nullable Iterable iterable) =>
      iterable == null || iterable.isEmpty;
}

@beta
Iterable<int> lazySynchronousNaturalsTo(int n) sync* {
  int k = 0;
  while (k < n) {
    yield k++;
  }
}

@beta
Stream<int> lazyAsynchronousNaturalsTo(int n) async* {
  int k = 0;
  while (k < n) {
    yield k++;
  }
}

extension IterableX<T> on Iterable<T> {

  /// Returns `true` if this nullable collection is either null or empty
  bool get isNullOrEmpty => this == null || this.isEmpty;

  /// Returns an element at the given [index] or `null` if the [index]
  /// is out of bounds of this collection.
  @nullable
  T elementAtOrNull(int index) {
    if (this is List) {
      return ListX(this as List).getOrNull(index);
    }
    if (index < 0) return null;
    var iterator = this.iterator;
    var count = 0;
    while (iterator.moveNext()) {
      var element = iterator.current;
      if (index == count++) {
        return element;
      }
    }
    return null;
  }

  /// Returns the first element matching the given [predicate], or `null` if element was not found.
  T firstOrNull(bool predicate(T t)) {
    for (var element in this) {
      if (predicate(element)) {
        return element;
      }
    }
    return null;
  }

  /// Returns `true` if all elements match the given [predicate].
  bool all(bool predicate(T t)) {
    if (this.isEmpty) return true;
    for (var element in this) {
      if (!predicate(element)) return false;
    }
    return true;
  }

  /// Groups elements of the original collection by the key returned by the
  /// given [keySelector] function applied to each element and returns a map
  /// where each group key is associated with a list of corresponding elements.
  ///
  /// The returned map preserves the entry iteration order of the keys produced
  /// from the original collection.
  @nonNull
  Map<K, List<T>> groupBy<K>(K keySelector(T t)) {
    return IterableX<T>(this).groupByTo<K, Map<K, List<T>>>(
        LinkedHashMap<K, List<T>>(), keySelector);
  }

  /// Groups values returned by the [valueTransform] function applied to each element of the original collection
  /// by the key returned by the given [keySelector] function applied to the element
  /// and returns a map where each group key is associated with a list of corresponding values.
  ///
  /// The returned map preserves the entry iteration order of the keys produced from the original collection.
  Map<K, List<V>> groupBy2<K, V>(
      @nonNull K Function(T t) keySelector, @nonNull V valueTransform(T t)) {
    return IterableX(this).groupByTo2<K, V, Map<K, List<V>>>(
        LinkedHashMap<K, List<V>>(), keySelector, valueTransform);
  }

  /// Groups elements of the original collection by the key returned by the
  /// given [keySelector] function applied to each element and puts to
  /// the [destination] map each group key associated with a
  /// list of corresponding elements.
  @nonNull
  M groupByTo<K, M extends Map<K, List<T>>>(
      @nonNull M destination, @nonNull K keySelector(T t)) {
    checkNotNull(destination);
    checkNotNull(keySelector);
    for (var element in this) {
      var key = keySelector(element);
      var list = destination.getOrPut(key, () => <T>[]);
      list.add(element);
    }
    return destination;
  }

  /// Groups values returned by the [valueTransform] function applied to each
  /// element of the original collection by the key returned by the
  /// given [keySelector] function applied to the element and puts to
  /// the [destination] map each group key associated with a list of corresponding values.
  M groupByTo2<K, V, M extends Map<K, List<V>>>(@nonNull M destination,
      @nonNull K keySelector(T t), @nonNull V valueTransform(T t)) {
    checkNotNull(destination);
    checkNotNull(keySelector);
    for (var element in this) {
      var key = keySelector(element);
      var list = destination.getOrPut(key, () => <V>[]);
      list.add(valueTransform(element));
    }
    return destination;
  }
}

extension ListX<T> on List<T> {
  /// Returns the index of the last item in the list or -1 if the list is empty
  int get lastIndex => this.length - 1;

  /// Returns an element at the given [index] or `null` if the [index]
  /// is out of bounds of this list.
  @nullable
  T getOrNull(int index) {
    if (index >= 0 && index <= ListX(this).lastIndex) {
      return this[index];
    } else {
      return null;
    }
  }

  /// Returns an element at the given [index] or the result of calling
  /// the [defaultValue] function if the [index] is out of bounds of this list.
  @nonNull
  T elementAtOrElse(int index, T Function(int index) defaultValue) {
    return ListX(this).getOrNull(index) ?? defaultValue(index);
  }
}
